/* Copyright (c) 2013 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *	 notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *	 notice, this list of conditions and the following disclaimer in the
 *	 documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *	 names of its contributors may be used to endorse or promote products
 *	 derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <internal/process.h>
#include "dpa_sys.h"


struct process_interrupt {
	int irq;
	irqreturn_t (*isr)(int irq, void *arg);
	unsigned long flags;
	const char *name;
	void *arg;
	struct list_head node;
};

static LIST_HEAD(process_irq_list);
static pthread_mutex_t process_irq_lock = PTHREAD_MUTEX_INITIALIZER;

void process_interrupt_install(struct process_interrupt *irq)
{
	/* Add the irq to the end of the list */
	pthread_mutex_lock(&process_irq_lock);
	list_add_tail(&irq->node, &process_irq_list);
	pthread_mutex_unlock(&process_irq_lock);
}
void process_interrupt_remove(struct process_interrupt *irq)
{
	pthread_mutex_lock(&process_irq_lock);
	list_del(&irq->node);
	pthread_mutex_unlock(&process_irq_lock);
}
struct process_interrupt *process_interrupt_find(int irq_num)
{
	struct process_interrupt *i = NULL;
	pthread_mutex_lock(&process_irq_lock);
	list_for_each_entry(i, &process_irq_list, node) {
		if (i->irq == irq_num)
			goto done;
	}
done:
	pthread_mutex_unlock(&process_irq_lock);
	return i;
}


/* This is the interface from the platform-agnostic driver code to (de)register
 * interrupt handlers. We simply create/destroy corresponding structs. */
int qbman_request_irq(int irq, irqreturn_t (*isr)(int irq, void *arg),
			unsigned long flags, const char *name, void *arg)
{
	struct process_interrupt *irq_node =
		kmalloc(sizeof(*irq_node), GFP_KERNEL);

	if (!irq_node)
		return -ENOMEM;
	irq_node->irq = irq;
	irq_node->isr = isr;
	irq_node->flags = flags;
	irq_node->name = name;
	irq_node->arg = arg;
	process_interrupt_install(irq_node);
	return 0;
}

int qbman_free_irq(int irq, void *arg)
{
	struct process_interrupt *irq_node = process_interrupt_find(irq);
	if (!irq_node)
		return -EINVAL;
	process_interrupt_remove(irq_node);
	kfree(irq_node);
	return 0;
}
/* This is the interface from the platform-specific driver code to obtain
 * interrupt handlers that have been registered. */
void qbman_invoke_irq(int irq)
{
	struct process_interrupt *irq_node = process_interrupt_find(irq);
	if (irq_node)
		irq_node->isr(irq, irq_node->arg);

}
